"""
 DEX SWAP
✍️ CONTRACT
"""

from datetime import datetime

import requests
from loguru import logger
from pycoingecko import CoinGeckoAPI

from dxsp.config import settings
from dxsp.utils.utils import fetch_url


class ContractUtils:
    """
    Contract Utils used to interact with the dex protocol

    Args:
        w3 (Optional[Web3]): Web3
        block_explorer_url (Optional[str]): block_explorer_url
        block_explorer_api (Optional[str]): block_explorer_api

    Returns:
        ContractUtils

    Methods:
        get_data()
        search()
        get_cg_platform()
        get_tokenlist_data()
        get_cg_data()
        get_confirmation()

    """

    def __init__(self, w3=None, block_explorer_url=None, block_explorer_api=None):
        """
        Initializes an instance of the class.

        :param w3: An instance of the web3.py library.
        :type w3: web3.Web3

        :param block_explorer_url: The URL of the block explorer.
        :type block_explorer_url: str

        :param block_explorer_api: The API endpoint of the block explorer.
        :type block_explorer_api: str
        """
        self.w3 = w3
        logger.debug("w3: {}", self.w3)
        self.chain = int(self.w3.net.version, 16)
        logger.debug("chain: {}", self.chain)
        self.block_explorer_url = block_explorer_url
        self.block_explorer_api = block_explorer_api
        self.cg = CoinGeckoAPI()
        self.platform = self.get_cg_platform()
        logger.debug("platform: {}", self.platform)

    async def get_data(self, symbol=None, contract_address=None):
        """
        Get data based on the provided symbol or contract address.

        Args:
            symbol (str): The symbol to search for.
            contract_address (str): The contract address of the token.

        Returns:
            Token: The token object containing the data if contract_address is provided.
            None: If neither symbol nor contract_address is provided.
        """
        if symbol:
            return await self.search(symbol)
        if contract_address:
            token = Token(
                w3=self.w3,
                address=contract_address,
            )
            await token.fetch_data()
            return token
        return None

    def get_cg_platform(self):
        """
        Retrieves the platform associated
        with the current network.

        Returns:
            str: The coingecko platform name of the platform associated
            with the current network, or None if no platform is found.

        Raises:
            Exception: If an error occurs
             while retrieving the platform.

        """

        logger.debug("get_cg_platform")
        logger.debug("chain: {}", self.chain)
        network_versions = {
            1: "ethereum",
            10: "optimistic-ethereum",
            56: "binance-smart-chain",
            137: "polygon-pos",
            250: "fantom",
            43114: "avalanche",
            42161: "arbitrum-one",
        }
        if network_name := network_versions.get(self.chain):
            logger.debug("coingecko platform identified {}", network_name)
            return network_name
        try:
            asset_platforms = self.cg.get_asset_platforms()
            output_dict = next(
                x for x in asset_platforms if x["chain_identifier"] == self.chain
            )
            platform = output_dict["id"] or None
            logger.debug("coingecko platform identified {}", platform)
            return platform
        except StopIteration:
            logger.error("No matching platform found for chain: {}", self.chain)
            return None
        except Exception as e:
            logger.error("get_cg_platform: {}", e)
            return None

    async def search(self, token):
        """
        Asynchronously searches for a token based on the given token parameter.

        Args:
            token (str): The token to search for.

        Returns:
            Token: An instance of the Token class representing the found token.

        Raises:
            Exception: If the token is not found.
        """
        try:
            token_instance = await self.search_tokenlist_data(token)
            if token_instance is None:
                logger.info("Searching on Coingecko")
                token_instance = await self.search_cg_data(token)

            if token_instance is None:
                raise Exception(f"Token not found: {token} on {self.chain}")
            token_instance.block_explorer_api = self.block_explorer_api
            token_instance.block_explorer_url = self.block_explorer_url
            return token_instance
        except Exception as e:
            logger.error("Search {}: {}", token, e)
            raise

    async def search_tokenlist_data(self, token):
        """
        Asynchronously searches for tokenlist data based on a given token.

        :param token: The token to search for.
        :type token: str

        :return: An instance of the Token class
        if the tokenlist data is found, else None.
        :rtype: Token or None
        """
        result = await self.get_tokenlist_data(token)
        if result is not None:
            token_instance = Token(w3=self.w3, address=result["address"])
            token_instance.decimals = result["decimals"]
            token_instance.symbol = result["symbol"]
            return token_instance
        return None

    async def get_tokenlist_data(self, symbol):
        """
        Retrieves token data from a given token list URL based on the provided symbol.

        Args:
            symbol (str): The symbol of the token to search for.

        Returns:
            dict or None: The token data if found, None otherwise.
        """
        try:
            token_list_urls = [
                settings.token_personal_list,
                settings.token_testnet_list,
                settings.token_mainnet_list,
            ]

            for token_list_url in token_list_urls:
                if not token_list_url:
                    continue
                logger.debug("Token search in {}", token_list_url)
                token_list = await fetch_url(token_list_url)
                token_search = token_list["tokens"]
                for keyval in token_search:
                    if keyval["symbol"] == symbol and keyval["chainId"] == self.chain:
                        logger.debug("token data found {}", keyval)
                        return keyval
                    elif keyval["address"] == symbol and keyval["chainId"] == 1:
                        logger.debug("token data found {}", keyval)
                        return keyval
                logger.warning(f"Token {symbol} not found on list")
        except Exception as e:
            logger.error("get_token_data: {}", e)
            return None

    async def search_cg_data(self, token):
        """
        Asynchronously searches for CG data using the provided token.

        :param token: The token to search for.
        :type token: Any

        :return: The token instance if found on Coingecko, otherwise None.
        :rtype: Optional[Token]
        """
        result = await self.get_cg_data(token)
        if result is not None:
            logger.info("Found on Coingecko")
            token_instance = Token(w3=self.w3, address=result["contract_address"])
            token_instance.decimals = result["decimal_place"]
            return token_instance
        return None

    async def get_cg_data(self, token):
        """
        Retrieves data for a given token from the CoinGecko API.

        Args:
            token (str): The symbol of the token to retrieve data for.

        Returns:
            str or None: The data for the token on the specified platform,
                         or None if the token is not found or an error occurs.
        """
        # todo: add support for address search
        try:
            if self.platform is None:
                return None
            search_results = self.cg.search(query=token)
            search_dict = search_results["coins"]
            # logger.debug("Coingecko search results: {}", search_dict)
            filtered_dict = [x for x in search_dict if x["symbol"] == token.upper()]
            api_dict = [sub["api_symbol"] for sub in filtered_dict]
            for i in api_dict:
                coin_dict = self.cg.get_coin_by_id(i)
                try:
                    if coin_dict["detail_platforms"][f"{self.platform}"]:
                        return coin_dict["detail_platforms"][f"{self.platform}"]
                except (KeyError, requests.exceptions.HTTPError):
                    pass
        except Exception as e:
            logger.error("get_cg_data {}", e)

    async def get_confirmation(self, transaction_hash):
        """

        Returns trade confirmation.

        Args:
            transaction_hash (str): The transaction hash

        Returns:
            dict: The trade confirmation

        Raises:
            Exception: Error

        """
        try:
            transaction = self.w3.eth.get_transaction(transaction_hash)
            block_info = self.w3.eth.get_block(transaction["blockNumber"])
            return {
                "timestamp": datetime.utcfromtimestamp(block_info["timestamp"]),
                "id": transaction_hash,
                "instrument": transaction["to"],
                "amount": transaction["value"],
                "price": transaction["gasPrice"],
                "fee": block_info["gasUsed"],
                "confirmation": (
                    f"➕ Size: {round(transaction['value'], 4)}\n"
                    f"ℹ️ {transaction_hash}\n"
                    f"⛽ {block_info['gasUsed']}\n"
                    f"🗓️ {datetime.utcfromtimestamp(block_info['timestamp'])}"
                ),
            }
        except Exception as error:
            logger.error("get_confirmation {}", error)


class Token:
    """

    Class Token to interact with web3 token contract

    Args:
        w3: An instance of the web3 library.
        address: The address of the token contract.
        block_explorer_url: The URL of the block explorer for the token.
        block_explorer_api: The API endpoint of the block explorer for the token.
        symbol: The symbol of the token.

    Returns:
        Token

    Methods:
        __init__: Initializes an instance of the class.
        fetch_data: Retrieves data for the token.
        get_token_abi: Retrieves the token abi.
        get_token_contract: Retrieves the token contract.
        get_contract_function: Retrieves the contract functions by name.
        get_token_balance: Retrieves the token balance.
        get_token_symbol: Retrieves the token symbol.
        get_token_name: Retrieves the token name.
        get_token_decimals: Retrieves the token decimals.

    """

    def __init__(
        self,
        w3=None,
        address=None,
        block_explorer_url=None,
        block_explorer_api=None,
        symbol=None,
    ):
        """
        Initializes an instance of the class.

        :param w3: An instance of the web3 library.
        :type w3: object

        :param address: The address of the token contract.
        :type address: str

        :param block_explorer_url: The URL of the block explorer for the token.
        :type block_explorer_url: str

        :param block_explorer_api: The API endpoint of the block explorer for the token.
        :type block_explorer_api: str

        :param symbol: The symbol of the token.
        :type symbol: str
        """
        try:
            self.w3 = w3
            self.address = self.w3.to_checksum_address(address)
            self.symbol = symbol
            self.block_explorer_url = block_explorer_url
            self.block_explorer_api = block_explorer_api
            self.decimals = None
            self.name = None
            logger.debug(
                "token initialized symbol {} address {}", self.symbol, self.address
            )
        except Exception as error:
            logger.error("token error {}", error)

    async def fetch_data(self) -> None:
        """
        Retrieves data for the token.
        """
        logger.debug("fetch token data")
        self.contract = await self.get_token_contract()
        self.decimals = await self.get_token_decimals()
        self.symbol = await self.get_token_symbol()
        self.name = await self.get_token_name()
        logger.debug("{} - token data {}", self.symbol, self.address)

    async def get_token_abi(self, address=None):
        """
        Retrieves the ABI (Application Binary Interface)
        of a token contract at the given address.

        Args:
            address (str, optional): The address of the token contract.
            If not provided, the address associated with the
             contract instance is used. Defaults to None.

        Returns:
            str: The ABI of the token contract, if successful.
            None if the request fails or the contract does not have an ABI.
        """
        if not self.block_explorer_api:
            return await fetch_url(settings.dex_erc20_abi_url)
        if address is None:
            address = self.address
        params = {
            "module": "contract",
            "action": "getabi",
            "address": address,
            "apikey": self.block_explorer_api,
        }
        resp = await fetch_url(
            url=self.block_explorer_url, headers=settings.headers, params=params
        )
        if resp:
            return resp["result"] if resp["status"] == "1" else None

    async def get_token_contract(self):
        """
        Retrieves the token contract.

        :return: The token contract.
        """
        self.abi = await self.get_token_abi()
        if self.abi is None:
            return None
        contract = self.w3.eth.contract(address=self.address, abi=self.abi)
        return contract

    def get_contract_function(self, contract, func_name: str):
        """
        Get the contract function by name.

        Args:
            contract: The contract object.
            func_name (str): The name of the function.

        Returns:
            bool: True if the function exists
            in the contract, False otherwise.
        """
        return func_name in dir(contract.functions)

    async def get_token_balance(self, wallet_address):
        """
        Get the balance of a token for a given wallet address.

        Args:
            wallet_address (str): The wallet
            address to check the balance for.

        Returns:
            float: The balance of the token in ether.

        Raises:
            None
        """
        contract = await self.get_token_contract()
        if contract is None or contract.functions is None:
            logger.warning("No Balance")
            return 0
        balance = contract.functions.balanceOf(wallet_address).call()
        if balance is None:
            logger.warning("No Balance")
            return 0
        return round(self.w3.from_wei(balance, "ether"), 5) or 0

    async def get_token_symbol(self):
        """
        Retrieves the symbol of the token.

        Returns:
            str: The symbol of the token.
        """
        contract = await self.get_token_contract()
        return contract.functions.symbol().call()

    async def get_token_name(self):
        """
        Get the name of the token.

        Args:
            self: The object itself.

        Returns:
            The name of the token as a string.
        """
        contract = await self.get_token_contract()
        return contract.functions.name().call()

    async def get_token_decimals(self):
        """
        Get the number of decimal places for the token.

        Returns:
            int: The number of decimal places for the token.
        """
        if self.decimals is not None:
            return self.decimals
        contract = await self.get_token_contract()
        return contract.functions.decimals().call() if contract else 18



"""
Base DexClient Class   🦄
"""

import decimal
from datetime import datetime, timedelta

import aiohttp
from loguru import logger
from web3 import Web3
from web3.gas_strategies.time_based import medium_gas_price_strategy
from web3.middleware import geth_poa_middleware

from dxsp.utils import AccountUtils, ContractUtils


class DexClient:
    """
    Base DexClient Class

    Args:
        **kwargs:

    Returns:
        None

    Methods:
        resolve_buy_token
        resolve_sell_token
        resolve_token
        replace_instrument
        get_order_amount
        get_quote
        get_swap
        make_swap
        get_account_balance
        get_trading_asset_balance
        get_account_position
        get_account_margin
        get_account_open_positions
        get_account_pnl
        calculate_pnl

    """

    def __init__(self, **kwargs):

        self.name = kwargs.get("name", None)
        logger.debug(f"Setting up: {self.name}")

        self.protocol = kwargs.get("protocol") or "uniswap"
        self.protocol_version = kwargs.get("protocol_version") or 2
        self.api_endpoint = kwargs.get("api_endpoint", None)
        self.api_key = kwargs.get("api_key", None)
        self.rpc = kwargs.get("rpc", None)
        self.w3 = Web3(Web3.HTTPProvider(self.rpc))
        if self.w3:
            self.w3.middleware_onion.inject(geth_poa_middleware, layer=0)
            self.w3.eth.set_gas_price_strategy(medium_gas_price_strategy)
            logger.debug("Chain hex {}", self.w3.net.version)
            logger.debug("Chain {}", int(self.w3.net.version, 16))

        self.wallet_address = kwargs.get("wallet_address", None)
        self.private_key = kwargs.get("private_key", None)
        if self.w3 and self.wallet_address:
            self.account_number = (
                f"{int(self.w3.net.version, 16)} - {str(self.wallet_address)[-8:]}"
            )
        else:
            self.account_number = None
        logger.debug("Account {}", self.account_number)

        self.router_contract_addr = kwargs.get("router_contract_addr", None)
        self.factory_contract_addr = kwargs.get("factory_contract_addr", None)
        self.trading_asset_address = kwargs.get("trading_asset_address", None)
        self.trading_risk_percentage = kwargs.get("trading_risk_percentage", None)
        self.trading_asset_separator = kwargs.get("trading_asset_separator", None)
        self.trading_risk_amount = kwargs.get("trading_risk_amount", None)
        self.trading_slippage = kwargs.get("trading_slippage", None)
        self.trading_amount_threshold = kwargs.get("trading_amount_threshold", None)
        self.block_explorer_url = kwargs.get("block_explorer_url", None)
        self.block_explorer_api = kwargs.get("block_explorer_api", None)
        self.mapping = kwargs.get("mapping", None)
        self.is_pnl_active = kwargs.get("is_pnl_active", False)
        self.rotki_report_endpoint = kwargs.get("rotki_report_endpoint", None)

        self.contract_utils = ContractUtils(
            self.w3, self.block_explorer_url, self.block_explorer_api
        )
        self.account = AccountUtils(
            self.w3,
            self.contract_utils,
            self.wallet_address,
            self.private_key,
            self.trading_asset_address,
            self.block_explorer_url,
            self.block_explorer_api,
        )
        self.client = None

    async def resolve_token(self, **kwargs):
        """
        A function to resolve a token based on the input address or symbol.
        It takes *args and **kwargs as input parameters.
        Returns the data associated with the token.

        Args:
            **kwargs: either an address or a symbol.

        Returns:
            Token: The token object containing the data if contract_address is provided.
            None: If neither symbol nor contract_address is provided.
        """
        logger.debug("Resolving token {}", kwargs)
        try:
            (identifier,) = kwargs.values()
        except ValueError as e:
            raise ValueError(
                "Token identification must be an address or a symbol"
            ) from e

        # Determine if the input is an address or a symbol
        # Assuming addresses start with '0x'
        if identifier.startswith("0x"):
            result = await self.contract_utils.get_data(contract_address=identifier)
        else:
            symbol = await self.replace_instrument(identifier)
            result = await self.contract_utils.get_data(symbol=symbol)

        # Check if the result is not None
        if not result:
            raise ValueError("Token {} not found", identifier)

        return result

    async def replace_instrument(self, instrument):
        """
        Replace instrument by an alternative instrument, if the
        instrument is not in the mapping, it will be ignored.
        Mapping, define in settings as TOML or .env variable.
        It is a list of dictionaries such as:
        mapping = [
            { id = "BTC", alt = "WBTC" ,enable = true },
        ]

        Args:
            instrument (str):

        Returns:
            dict
        """
        logger.debug("Replace instrument: {}", instrument)
        if self.mapping is None:
            return instrument
        for item in self.mapping:
            if item["id"] == instrument:  # and item["enable"] is not False:
                instrument = item["alt"]
                logger.debug("Instrument symbol changed {}", instrument)
                break
        logger.debug("Instrument symbol changed {}", instrument)
        return instrument

    async def get_order_amount(
        self, sell_token, wallet_address, quantity, is_percentage=True
    ):
        """
        Calculate the order amount based on the sell token,
        wallet address, quantity, and whether it is a percentage.

        Args:
            sell_token (SellToken): The sell token object.
            wallet_address (str): The wallet address.
            quantity (float): The quantity of the sell token.
            is_percentage (bool, optional):
            Flag indicating whether the quantity is a percentage. Defaults to True.

        Returns:
            float: The calculated order amount.
        """
        logger.debug("get order amount {} {} {}", sell_token, wallet_address, quantity)
        logger.debug("Protocol", self.contract_utils.platform)
        balance = await sell_token.get_token_balance(wallet_address)
        logger.debug("Balance {}", balance)
        if not is_percentage and balance:
            logger.debug("Quantity {}", quantity)
            return quantity

        if balance:
            risk_percentage = float(quantity) / 100
            logger.debug("Risk percentage {}", risk_percentage)
            amount = balance * decimal.Decimal(risk_percentage)
            logger.debug("Amount {}", amount)
            if (
                isinstance(amount, decimal.Decimal)
                and amount > self.trading_amount_threshold
            ):
                logger.debug("Amount {}", amount)
                return amount

        return 0

    async def get_quote(
        self,
        buy_address=None,
        buy_symbol=None,
        sell_address=None,
        sell_symbol=None,
        amount=1,
    ):
        """
        Get a quote method for specific protocol

        """

    async def get_swap(self, sell_token=None, buy_token=None, quantity=1):
        """
        Execute a swap

        Args:
            sell_token (str): The sell token.
            buy_token (str): The buy token.
            quantity (int): The quantity of tokens.

        Returns:
            transactionHash


        """
        try:
            logger.debug("get swap {} {} {}", sell_token, buy_token, quantity)
            logger.debug("Protocol", self.contract_utils.platform)
            sell_token = await self.contract_utils.get_data(symbol=sell_token)
            logger.debug("sell token {}", sell_token)
            buy_token = await self.contract_utils.get_data(symbol=buy_token)
            logger.debug("buy token {}", buy_token)

            sell_amount = await self.get_order_amount(
                sell_token, self.account.wallet_address, quantity
            )
            if not sell_amount or sell_amount == 0:
                logger.error("sell amount {}", sell_amount)
                return f"⚠️ sell amount failed {sell_amount}"

            sell_token_amount_wei = decimal.Decimal(sell_amount) * (
                decimal.Decimal("10") ** int(sell_token.decimals)
            )
            if self.protocol == "0x":
                await self.account.get_approve(sell_token.address)

            order_amount = int(
                sell_token_amount_wei * decimal.Decimal((self.trading_slippage / 100))
            )
            logger.debug("order amount {}", order_amount)

            order = await self.make_swap(
                sell_token.address, buy_token.address, order_amount
            )

            if not order:
                logger.error("swap order not executed")
                return "⚠️ order execution failed"

            signed_order = await self.account.get_sign(order)
            order_hash = str(self.w3.to_hex(signed_order))
            logger.debug(order_hash)
            receipt = self.w3.eth.wait_for_transaction_receipt(order_hash)
            logger.debug(receipt)

            if receipt["status"] != 1:
                logger.error("receipt failed")

            return await self.contract_utils.get_confirmation(
                receipt["transactionHash"]
            )

        except Exception as error:
            logger.debug(error)
            return "⚠️ " + str(error)

    async def make_swap(self, sell_address, buy_address, amount):
        """
        Make a swap method for specific protocol

        """

    async def get_account_balance(self):
        """
        Retrieves the account balance.

        :return: The account balance.
        :rtype: float
        """
        return await self.account.get_account_balance()

    async def get_trading_asset_balance(self):
        """
        Retrieves the trading asset balance for the current account.

        :return: A dictionary containing the trading asset balance.
                 The dictionary has the following keys:
                 - 'asset': The asset symbol.
                 - 'free': The free balance of the asset.
                 - 'locked': The locked balance of the asset.
        """
        return await self.account.get_trading_asset_balance()

    async def get_account_position(self):
        """
        Retrieves the account position.

        :return: The account position.
        :rtype: AccountPosition
        """
        return await self.account.get_account_position()

    async def get_account_margin(self):
        """
        Retrieves the account margin.

        :return: The account margin.
        :rtype: float
        """
        return await self.account.get_account_margin()

    async def get_account_open_positions(self):
        """
        Retrieves the open positions of the account.

        :return: A list of open positions in the account.
        """
        return await self.account.get_account_open_positions()

    async def get_account_pnl(self, period=None):
        """
        Return account pnl.

        Args:
            None

        Returns:
            pnl
        """
        today = datetime.now().date()
        if period is None:
            start_date = today
        elif period == "W":
            start_date = today - timedelta(days=today.weekday())
        elif period == "M":
            start_date = today.replace(day=1)
        elif period == "Y":
            start_date = today.replace(month=1, day=1)
        else:
            return 0
        return self.calculate_pnl(start_date) if self.is_pnl_active else 0

    async def calculate_pnl(self, period=None):
        """
        Calculate the PnL for a given period.
        via https://rotki.readthedocs.io/en/latest/api.html

        Parameters:
            period (str):
            The period for which to calculate PnL ('W', 'M', 'Y', or None)

        Returns:
            pnl: The calculated PnL value.
        """
        if self.rotki_report_endpoint is None:
            return 0
        params = {"period": period} if period else {}

        async with aiohttp.ClientSession() as session:
            async with session.get(
                self.rotki_report_endpoint, params=params
            ) as response:
                if response.status != 200:
                    logger.error(f"Received non-200 status code: {response.status}")
                    return 0
                data = await response.json()
                result = data.get("result", {})
                entries = result.get("entries", [])
                # Initialize a dictionary to hold the sum of 'free' values
                free_values = {
                    "trade": 0,
                    "transaction event": 0,
                    "fee": 0,
                    "asset movement": 0,
                }
                for entry in entries:
                    overview = entry.get("overview", {})
                    for category, amounts in overview.items():
                        try:
                            free_amount = float(amounts.get("free", "0"))
                            # Add it to the total
                            free_values[category] += free_amount
                        except ValueError:
                            logger.error(f"Invalid free amount: {amounts.get('free')}")

                return free_values
